import os
import string
import StringIO
import random
import math
import time
import sys
import socket
import signal
import pickle

if os.environ.has_key('DISPLAY'):
    if not sys.modules.has_key('gtk'):
        try:
            import pygtk
            pygtk.require("2.0")
        except ImportError:
            pass
    import gtk
    import pango
else:
    gtk = None

import ninix.pix
import ninix.script

import kanjilib

def read_bln_txt(dir):
    blns = {}
    try:
        file = open(os.path.join(dir, 'bln.txt'), 'r')
        data = {}
        name = ''
        while 1:
            line = file.readline()
            if not line:
                if name:
                    blns[name] = (data, {})
                break # EOF
            line = string.strip(line)
            if not line:
                continue
            if line[:2] == '//':
                continue
            start = string.find(line, '[')
            if start >= 0:
                if name:
                    blns[name] = (data, {})
                data = {}
                end = string.find(line, ']')
                if end < 0:
                    end = len(line)
                name = line[start+1:end]
            else:
                comma = string.find(line, ',')
                if comma >= 0:
                    key = string.strip(line[:comma])
                    value = string.strip(line[comma+1:])
                    data[key] = value
        return blns
    except:
        return {}

class Saori:
    def __init__(self):
	self.loaded = 0
        self.blns = {}
        self.sakura = None
        self.timeout_id = None
    def use_sakura(self, sakura):
        self.sakura = sakura
    def load(self, dir=os.curdir):
        self.dir = dir
        result = 0
        if not self.sakura or not gtk:
            pass
        elif self.loaded:
	    result = 2
        else:
            self.blns = read_bln_txt(self.dir)
            if self.blns:
                self.talking = 0
                self.timeout_id = None
                self.loaded = 1
                result = 1
	return result
    def unload(self):
        for name in self.blns.keys():
            data, bln = self.blns[name]
            for bln_id in bln.keys():
                if bln[bln_id][0]:
                    try:
                        os.kill(-bln[bln_id][0], signal.SIGTERM)
                    except:
                        pass
                    del bln[bln_id]
        self.blns = {}
        if self.timeout_id:
            gtk.timeout_remove(self.timeout_id)
            self.timeout_id = None
        self.loaded = 0
        return 1
    def request(self, req):
        type, argument = self.evaluate_request(req)
        header = StringIO.StringIO(req)
        line = header.readline()
        if not type:
            return 'SAORI/1.0 400 Bad Request\r\n\r\n'
        elif type == 'GET Version':
            return 'SAORI/1.0 204 No Content\r\n\r\n'
        elif type == 'EXECUTE':
            if len(argument) == 0:
                return 'SAORI/1.0 400 Bad Request\r\n\r\n'
            name = argument[0]
            if len(argument) >= 2:
                text = argument[1]
            else:
                text = ''
            if len(argument) >= 3:
                offset_x = int(argument[2])
            else:
                offset_x = 0
            if len(argument) >= 4:
                offset_y = int(argument[3])
            else:
                offset_y = 0
            if len(argument) >= 5:
                bln_id = argument[4]
            else:
                bln_id = ''
            if self.blns.has_key(name):
                data, bln = self.blns[name]
                if bln.has_key(bln_id):
                    try:
                        os.kill(-bln[bln_id][0], signal.SIGTERM)
                    except:
                        pass
                    del bln[bln_id]
                if text:
                    life_time = None
                    if data.has_key('life'):
                        life = data['life']
                        if life == 'auto':
                            life_time = 16000
                        elif life == 'infinitely' or life == 'orusuban':
                            pass
                        else:
                            try:
                                life_time = int(life)
                            except:
                                pass
                    else:
                        life_time = 16000
                    start_time = time.time()
                    bln[bln_id] = (None, text, life_time, start_time, offset_x, offset_y)
                self.blns[name] = (data, bln)
            elif name == 'clear':
                for name in self.blns.keys():
                    data, bln = self.blns[name]
                    if len(argument) >= 5:
                        if bln.has_key(bln_id) and \
                           not (data.has_key('life') and \
                                data['life'] == 'orusuban'):
                            try:
                                os.kill(-bln[bln_id][0], signal.SIGTERM)
                            except:
                                pass
                            del bln[bln_id]
                            self.blns[name] = (data, {})
                    else:
                        for bln_id in bln.keys():
                            if not (data.has_key('life') and \
                                    data['life'] == 'orusuban'):
                                try:
                                    os.kill(-bln[bln_id][0], signal.SIGTERM)
                                except:
                                    pass
                                del bln[bln_id]
                                self.blns[name] = (data, {})
            if not self.timeout_id:
                self.do_idle_tasks()
            return 'SAORI/1.0 204 No Content\r\n\r\n'
    def exec_balloon(self, name, bln_id):
        data, bln = self.blns[name]
        pid ,text, life_time, start_time, offset_x, offset_y = bln[bln_id]
        path = os.path.join(os.environ['PYTHONPATH'], 'ninix/dll/bln.py')
        try:
            s0_x, s0_y = self.sakura.surface.get_position(0)
            s0_w, s0_h = self.sakura.surface.get_surface_size(0)
        except:
            s0_x, s0_y = 0, 0
            s0_w, s0_h = 0, 0
        try:
            s1_x, s1_y = self.sakura.surface.get_position(1)
            s1_w, s1_h = self.sakura.surface.get_surface_size(1)
        except:
            s1_x, s1_y = 0, 0
            s1_w, s1_h = 0, 0
        try:
            b0_x, b0_y = self.sakura.balloon.get_position(0)
            b0_w, b0_h = self.sakura.balloon.get_balloon_size(0)
        except:
            b0_x, b0_y = 0, 0
            b0_w, b0_h = 0, 0
        try:
            b1_x, b1_y = self.sakura.balloon.get_position(1)
            b1_w, b1_h = self.sakura.balloon.get_balloon_size(1)
        except:
            b1_x, b1_y = 0, 0
            b1_w, b1_h = 0, 0
        bottommargin = self.sakura.bottom_margin
        directsstp_path = self.sakura.get_directsstp_path()
        args = [self.dir, text, offset_x, offset_y, name, bln_id, data,
                s0_x, s0_y, s0_w, s0_h, s1_x, s1_y, s1_w, s1_h,
                b0_x, b0_y, b0_w, b0_h, b1_x, b1_y, b1_w, b1_h,
                bottommargin, directsstp_path]
        args = pickle.dumps(args)
        os.execv('/usr/bin/env', ['-', 'python', path, args])
    def evaluate_request(self, req):
        type = None
        argument = []
        header = StringIO.StringIO(req)
        line = header.readline()
        if not line:
            return type, argument
        if line[-1] == '\n':
            line = line[:-1]
        line = string.strip(line)
        if not line:
            return type, argument
        for request in ['EXECUTE', 'GET Version']:
            if line[:len(request)] == request:
                type = request
                break
        while 1:
            line = header.readline()
            if not line:
                break # EOF
            if line[-1] == '\n':
                line = line[:-1]
            line = string.strip(line)
            if not line:
                continue
            colon = string.find(line, ':')
            if colon >= 0:
                key = string.strip(line[:colon])
                value = string.strip(line[colon+1:])
                if key[:8] == 'Argument':
                    argument.append(value)
                else:
                    continue
        return type, argument
    def kill_normal_balloons(self):
        for name in self.blns.keys():
            data, bln = self.blns[name]
            for bln_id in bln.keys():
                if not (data.has_key('life') and \
                        data['life'] == 'orusuban'):
                    try:
                        os.kill(-bln[bln_id][0], signal.SIGTERM)
                    except:
                        pass
                    del bln[bln_id]
                    self.blns[name] = (data, {})
    def do_idle_tasks(self):
        s0_shown, s0_x, s0_y, s0_w, s0_h = self.get_sakura_status('getSurfaceSakura')
        s1_shown, s1_x, s1_y, s1_w, s1_h = self.get_sakura_status('getSurfaceKero')
        b0_shown, b0_x, b0_y, b0_w, b0_h = self.get_sakura_status('getBalloonSakura')
        b1_shown, b1_x, b1_y, b1_w, b1_h = self.get_sakura_status('getBalloonKero')
        sakura_talking = self.get_sakura_status('isTalking')
        for name in self.blns.keys():
            data, bln = self.blns[name]
            for bln_id in bln.keys():
                pid, text, life_time, start_time, offset_x, offset_y = bln[bln_id]
                if data.has_key('life') and data['life'] == 'orusuban':
                    if pid:
                        if s0_shown or s1_shown:
                            try:
                                os.kill(-pid, signal.SIGTERM)
                            except:
                                pass
                            break
                    else:
                        if not s0_shown and not s1_shown:
                            gtk.gdk.flush()
                            self.kill_normal_balloons()
                            life_time = 300000
                            start_time = time.time()
                            self.pid = os.fork()
                            if self.pid == 0:
                                os.setsid()
                                self.pid = os.fork()
                                if self.pid == 0:
                                    self.exec_balloon(name, bln_id)
                                else:
                                    os._exit(1)
                            else:
                                os.waitpid(self.pid, 0)
                                bln[bln_id] = (self.pid, text, life_time, start_time, offset_x, offset_y)
                                time.sleep(5) # Important
                else:
                    if pid:
                        if not data.has_key('position'):
                            position = 'sakura'
                        else:
                            position = data['position']
                        if position == 'sakura' and not s0_shown:
                            try:
                                os.kill(-pid, signal.SIGTERM)
                            except:
                                pass
                            del bln[bln_id]
                            break
                        if position == 'kero' and not s1_shown:
                            try:
                                os.kill(-pid, signal.SIGTERM)
                            except:
                                pass
                            del bln[bln_id]
                            break
                        if position == 'sakurab' and not b0_shown:
                            try:
                                os.kill(-pid, signal.SIGTERM)
                            except:
                                pass
                            del bln[bln_id]
                            break
                        if position == 'kerob' and not b1_shown:
                            try:
                                os.kill(-pid, signal.SIGTERM)
                            except:
                                pass
                            del bln[bln_id]
                            break
                        nooverlap = 0
                        if data.has_key('nooverlap') and data['nooverlap'] == 'on':
                            nooverlap = 1
                        if nooverlap and \
                           not self.talking and sakura_talking:
                            try:
                                os.kill(-pid, signal.SIGTERM)
                            except:
                                pass
                            del bln[bln_id]
                            break
                    else:
                        if data.has_key('startdelay'):
                            startdelay = int(data['startdelay'])
                        else:
                            startdelay = 0
                        if time.time() - start_time >= startdelay * 0.001:
                            self.start_time = time.time()
                            self.pid = os.fork()
                            if self.pid == 0:
                                os.setsid()
                                self.pid = os.fork()
                                if self.pid == 0:
                                    self.exec_balloon(name, bln_id)
                                else:
                                    os._exit(1)
                            else:
                                os.waitpid(self.pid, 0)
                                bln[bln_id] = (self.pid, text, life_time, start_time, offset_x, offset_y)
                                # This is kluge.
                                self.sakura.script_wait = time.time() + 0.5
                if bln.has_key(bln_id):
                    pid, text, life_time, start_time, offset_x, offset_y = bln[bln_id]
                    if pid and life_time:
                        if time.time() - start_time >= life_time * 0.001:
                            try:
                                os.kill(-pid, signal.SIGTERM)
                            except:
                                pass
                            del bln[bln_id]
                            break
        if self.talking and not sakura_talking:
            self.talking = 0
        else:
            self.talking = sakura_talking
        self.timeout_id = gtk.timeout_add(5, self.do_idle_tasks)
    def get_sakura_status(self, id):
        if id == 'getSurfaceSakura':
            if self.sakura.surface.window and \
               self.sakura.surface.window[0].shown:
                s0_shown = 1
            else:
                s0_shown = 0
            try:
                s0_x, s0_y = self.sakura.surface.get_position(0)
                s0_w, s0_h = self.sakura.surface.get_surface_size(0)
            except:
                s0_x, s0_y = 0, 0
                s0_w, s0_h = 0, 0
            return s0_shown, s0_x, s0_y, s0_w, s0_h
        elif id == 'getSurfaceKero':
            if self.sakura.surface.window and \
               self.sakura.surface.window[1].shown:
                s1_shown = 1
            else:
                s1_shown = 0
            try:
                s1_x, s1_y = self.sakura.surface.get_position(1)
                s1_w, s1_h = self.sakura.surface.get_surface_size(1)
            except:
                s1_x, s1_y = 0, 0
                s1_w, s1_h = 0, 0
            return s1_shown, s1_x, s1_y, s1_w, s1_h
        elif id == 'getBalloonSakura':
            if self.sakura.balloon.window and \
               self.sakura.balloon.window[0].shown:
                b0_shown = 1
            else:
                b0_shown = 0
            try:
                b0_x, b0_y = self.sakura.balloon.get_position(0)
                b0_w, b0_h = self.sakura.balloon.get_balloon_size(0)
            except:
                b0_x, b0_y = 0, 0
                b0_w, b0_h = 0, 0
            return b0_shown, b0_x, b0_y, b0_w, b0_h
        elif id == 'getBalloonKero':
            if self.sakura.balloon.window and \
               self.sakura.balloon.window[1].shown:
                b1_shown = 1
            else:
                b1_shown = 0
            try:
                b1_x, b1_y = self.sakura.balloon.get_position(1)
                b1_w, b1_h = self.sakura.balloon.get_balloon_size(1)
            except:
                b1_x, b1_y = 0, 0
                b1_w, b1_h = 0, 0
            return b1_shown, b1_x, b1_y, b1_w, b1_h
        elif id == 'isTalking':
            talking = 0
            try:
                if self.sakura.processed_script or \
                   self.sakura.processed_text:
                    talking = 1
            except:
                pass
            return talking
        else:
            return None

class Balloon:
    def __init__(self, args):
        dir, text, offset_x, offset_y, name, bln_id, data, \
        s0_x, s0_y, s0_w, s0_h, s1_x, s1_y, s1_w, s1_h, \
        b0_x, b0_y, b0_w, b0_h, b1_x, b1_y, b1_w, b1_h, \
        bottommargin, directsstp_path = pickle.loads(args)
        signal.signal(signal.SIGTERM, self.signal_handler)
        self.dir = dir
        self.directsstp_path = directsstp_path
        self.dragged = 0
        self.name = name
        self.id = bln_id
        self.hup = 0
        self.layout = None
        if data.has_key('life') and data['life'] == 'orusuban':
            self.orusuban = 1
        else:
            self.orusuban = 0
        if not data.has_key('position'):
            self.position = 'sakura'
        else:
            self.position = data['position']
        self.window = gtk.Window(type=gtk.WINDOW_POPUP)
        self.window.connect("delete_event", self.delete)
        if data.has_key('clickerase') and data['clickerase'] == 'off':
            self.clickerase = 'off'
        else:
            self.clickerase = 'on'
        if data.has_key('dragmove.horizontal') and \
           data['dragmove.horizontal'] == 'on':
            self.dragmove_horizontal = 1
        else:
            self.dragmove_horizontal = 0
        if data.has_key('dragmove.vertical') and \
           data['dragmove.vertical'] == 'on':
            self.dragmove_vertical = 1
        else:
            self.dragmove_vertical = 0
        self.window.connect("button_press_event", self.button_press)
        self.window.connect("button_release_event", self.button_release)
        self.window.connect("motion_notify_event",  self.motion_notify)
        self.window.connect("leave_notify_event",  self.leave_notify)        
        self.window.set_events(gtk.gdk.BUTTON_PRESS_MASK|
                               gtk.gdk.BUTTON_RELEASE_MASK|
                               gtk.gdk.POINTER_MOTION_MASK|
                               gtk.gdk.LEAVE_NOTIFY_MASK)
        scrn_w = gtk.gdk.screen_width()
        scrn_h = gtk.gdk.screen_height()
        scrn_h = scrn_h - bottommargin
        if s0_x + s0_w/2 > scrn_w/2:
            self.direction0 = 0 # left
        else:
            self.direction0 = 1 # right
        if s1_x + s1_w/2 > scrn_w/2:
            self.direction1 = 0 # left
        else:
            self.direction1 = 1 # right            
        if ((self.position == 'sakura' and self.direction0) or \
            (self.position == 'kero' and self.direction1)) and \
            data.has_key('skin.right'):
            path = os.path.join(self.dir,
                                string.replace(data['skin.right'],
                                               '\\', '/'))
        elif (self.position == 'sakura' or self.position == 'kero') and \
             data.has_key('skin.left'):
            path = os.path.join(self.dir,
                                string.replace(data['skin.left'],
                                               '\\', '/'))
        elif data.has_key('skin'):
            path = os.path.join(self.dir,
                                string.replace(data['skin'],
                                               '\\', '/'))
        else:
            self.destroy()
        image, mask = ninix.pix.create_pixmap_from_file(path)
        self.x = 0
        self.y = 0
        w, h = image.get_size()
        if self.position == 'lefttop':
            pass
        elif self.position == 'leftbottom':
            self.y = scrn_h - h
        elif self.position == 'righttop':
            self.x = scrn_w - w
        elif self.position == 'rightbottom':
            self.x = scrn_w - w
            self.y = scrn_h - h
        elif self.position == 'center':
            self.x = (scrn_w - w)/2
            self.y = (scrn_h - h)/2
        elif self.position == 'sakura':
            if self.direction0: # right
                self.x = s0_x + s0_w
            else:
                self.x = s0_x - w
            self.y = s0_y
        elif self.position == 'kero':
            if self.direction1: # right
                self.x = s1_x + s1_w
            else:
                self.x = s1_x - w
            self.y = s1_y
        elif self.position == 'sakurab':
            self.x = b0_x
            self.y = b0_y
        elif self.position == 'kerob':
            self.x = b1_x
            self.y = b1_y
        if data.has_key('offset.x'):
            if (self.position == 'sakura' and not self.direction0) or \
               (self.position == 'kero' and not self.direction1):
                self.x = self.x - int(data['offset.x'])
            else:
                self.x = self.x + int(data['offset.x'])
        if data.has_key('offset.y'):
            self.y = self.y + int(data['offset.y'])
        if data.has_key('offset.random'):
            self.x = self.x + int(data['offset.random']) * random.randrange(-1, 2)
            self.y = self.y + int(data['offset.random']) * random.randrange(-1, 2)
        if (self.position == 'sakura' and not self.direction0) or \
           (self.position == 'kero' and not self.direction1):
            self.x = self.x - offset_x
        else:
            self.x = self.x + offset_x
        self.y = self.y + offset_y
        self.window.move(self.x, self.y)
        gtk.gdk.flush()
        self.left = 0
        self.right = w
        self.top = 0
        bottom = h
        if data.has_key('disparea.left'):
            self.left = int(data['disparea.left'])
        if data.has_key('disparea.right'):
            self.right = int(data['disparea.right'])
        if data.has_key('disparea.top'):
            self.top = int(data['disparea.top'])
        if data.has_key('disparea.bottom'):
            self.bottom = int(data['disparea.bottom'])
        self.mask_pixmap = mask
        self.window.shape_combine_mask(self.mask_pixmap, 0, 0)
        script = None
        self.darea = gtk.DrawingArea()
        self.darea.set_events(gtk.gdk.EXPOSURE_MASK)
        self.darea.connect("expose_event", self.redraw)
        self.darea.set_size_request(w, h)
        self.darea.show()
        self.window.add(self.darea)
        self.darea.realize()
        self.surface_gc = self.darea.window.new_gc()
        self.darea.window.set_back_pixmap(image, gtk.FALSE)
        if text != 'noscript' and (self.right-self.left) and (self.bottom-self.top):
            script = unicode(text, 'Shift_JIS', 'ignore')
            if data.has_key('font.color'):
                fontcolor_r = int(data['font.color'][:2], 16)
                fontcolor_g = int(data['font.color'][2:4], 16)
                fontcolor_b = int(data['font.color'][4:6], 16)
                cmap = self.darea.get_colormap()
                self.surface_gc.foreground = cmap.alloc_color(
                    "#%02x%02x%02x" % (fontcolor_r, fontcolor_g, fontcolor_b))
            self.font_size = 12 # default size
            if data.has_key('font.size'):
                self.font_size = int(data['font.size'])
            self.layout = pango.Layout(self.darea.get_pango_context())
            self.font_desc = pango.FontDescription()
            self.font_desc.set_family('Sans') # FIXME
            self.font_desc.set_size((self.font_size-3) * 1024) # size * PANGO_SCALE
            self.layout.set_font_description(self.font_desc)
            self.layout.set_wrap(pango.WRAP_CHAR)
            self.layout.set_width((self.right - self.left) * 1024)
        if data.has_key('slide.vx'):
            self.slide_vx = int(data['slide.vx'])
        else:
            self.slide_vx = 0
        if data.has_key('slide.vy'):
            self.slide_vy = int(data['slide.vy'])
        else:
            self.slide_vy = 0
        if data.has_key('slide.autostop'):
            self.slide_autostop = int(data['slide.autostop'])
        else:
            self.slide_autostop = 0
        if data.has_key('action.method') and \
           data['action.method'] == 'sinwave':
            action = data['action.method']
            ref0 = 0
            ref1 = 0
            ref2 = 0
            if data.has_key('action.reference0'):
                ref0 = int(data['action.reference0'])
            if data.has_key('action.reference1'):
                ref1 = int(data['action.reference1'])
            if data.has_key('action.reference2'):
                ref2 = int(data['action.reference2'])
            if ref1 and ref2:
                self.action = {'method': action,
                               'ref0': ref0, 'ref1': ref1, 'ref2': ref2}
            else:
                self.action = None
        else:
            self.action = None
        self.move_notify_time = None
        self.timeout_id = None
        self.visible = 0
        self.x_root = None
        self.y_root = None
        self.action_x = 0
        self.action_y = 0
        self.vx = 0
        self.vy = 0
        self.processed_script = None
        self.processed_text = ''
        self.text = ''
        self.script_wait = None
        self.script_parser = ninix.script.Parser(error="loose")
        try:
            self.processed_script = self.script_parser.parse(script)
        except ninix.script.ParserError, e:
            self.processed_script = None
            print "-" * 50
            print e
            print script
        self.window.show()
        self.start_time = time.time()
        self.timeout_id = gtk.timeout_add(10, self.do_idle_tasks)
        gtk.mainloop()
    def signal_handler(self, signum, frame):
        self.hup = 1
    def do_idle_tasks(self):
        if self.hup:
            self.destroy()
        if self.action and self.action['method'] == 'sinwave':
            offset = self.action['ref1'] \
                     * math.sin(2.0 * math.pi
                                * float(int((time.time() - self.start_time) * 1000) % self.action['ref2'])
                                / self.action['ref2'])
            if self.action['ref0']:
                self.action_y = offset
            else:
                self.action_x = offset
        if (self.slide_vx != 0 or self.slide_vy != 0) and \
               self.slide_autostop > 0 and \
               self.slide_autostop * 0.001 + 0.05 <= time.time() - self.start_time:
            self.vx = int((self.slide_autostop / 50.0 + 1) * self.slide_vx)
            if (self.position == 'sakura' and not self.direction0) or \
                   (self.position == 'kero' and not self.direction1):
                self.vx = -self.vx
            self.slide_vx = 0
            self.vy = int((self.slide_autostop / 50.0 + 1) * self.slide_vy)
            self.slide_vy = 0
        if self.slide_vx != 0:
            self.vx = int(((time.time() - self.start_time) * self.slide_vx) / 50 * 1000.0)
            if (self.position == 'sakura' and not self.direction0) or \
                   (self.position == 'kero' and not self.direction1):
                self.vx = -self.vx
        if self.slide_vy != 0:
            self.vy = int(((time.time() - self.start_time) * self.slide_vy) / 50 * 1000.0)
        self.window.move(self.x + self.action_x + self.vx,
                         self.y + self.action_y + self.vy)
        if self.processed_script or self.processed_text:
            self.interpret_script()
        self.timeout_id = gtk.timeout_add(10, self.do_idle_tasks)
    def redraw(self, widget, event):
        x, y, w, h = event.area
        self.darea.window.clear()
        if self.layout:
            self.darea.window.draw_layout(self.surface_gc, self.left, self.top, self.layout)
    def interpret_script(self):
        if self.script_wait is not None:
            if time.time() < self.script_wait:
                return
            self.script_wait = None
        if self.processed_text:
            self.text = self.text + self.processed_text[0]
            self.draw_text(self.text)
            self.processed_text = self.processed_text[1:]
            self.script_wait = time.time() + 0.014
            return
        node = self.processed_script.pop(0)
        if node[0] == ninix.script.SCRIPT_TAG:
            name, args = node[1], node[2:]
            if name == r'\n':
                self.text = self.text + '\n'
                self.draw_text(self.text)
            elif name == r'\w':
                try:
                    amount = int(args[0]) * 0.05 - 0.01
                except ValueError:
                    amount = 0
                if amount > 0:
                    self.script_wait = time.time() + amount
        elif node[0] == ninix.script.SCRIPT_TEXT:
            text = ''
            for chunk in node[1]:
                text = text + chunk[1]
            self.processed_text = text
    def draw_text(self, text):
        try:
            self.layout.set_text(text)
        except:
            self.layout.set_text(text, -1) # pygtk <= 1.99.14
        self.darea.queue_draw()
    def button_press(self, window, event):
        self.x_root = event.x_root
        self.y_root = event.y_root
        sstp_event = None
        if event.type == gtk.gdk.BUTTON_PRESS:
            if event.button == 1:
                sstp_event = 'OnEBMouseClick'
                ref = [self.name, int(event.x), int(event.y), self.id, 0]
            elif event.button == 3:
                sstp_event = 'OnEBMouseClick'
                ref = [self.name, int(event.x), int(event.y), self.id, 1]
        else:
            sstp_event = 'OnEBMouseDoubleClick'
            ref = [self.name, int(event.x), int(event.y), self.id]
        if sstp_event:
            header = "NOTIFY SSTP/1.0\r\n" + \
                     "Sender: easyballoon\r\n" + \
                     "Option: nodescript\r\n" + \
                     "Event: %s\r\n" % sstp_event
            for i in range(len(ref)):
                header = header + "Reference%d: %s\r\n" % (i, ref[i])
            header = header + "Charset: Shift_JIS\r\n\r\n"
            self.directsstp_send(header)
        if self.clickerase == 'on':
            self.destroy()
        return gtk.TRUE
    def directsstp_send(self, header):
        if not self.directsstp_path or self.orusuban:
            return None
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            s.connect(self.directsstp_path)
        except socket.error, (code, message):
            ##sys.stderr.write(message + "\n")
            return None
        s.send(header)
        buffer = []
        while 1:
            try:
                data = s.recv(1024)
            except socket.error, (code, message):
                continue
            if not data:
                break
            buffer.append(data)
        s.close()
        return kanjilib.sjis2euc(string.join(buffer, ''))
    def button_release(self, window, event):
        if self.dragged:
            self.dragged = 0
        self.x_root = None
        self.y_root = None
        return gtk.TRUE
    def motion_notify(self, window, event):
        if self.x_root is not None and \
           self.y_root is not None:
            x_delta = int(event.x_root - self.x_root)
            y_delta = int(event.y_root - self.y_root)
            if event.state & gtk.gdk.BUTTON1_MASK:
                self.dragged = 1
                if self.dragmove_horizontal:
                    self.x = self.x + x_delta
                if self.dragmove_vertical:
                    self.y = self.y + y_delta
                self.window.move(self.x + self.action_x + self.vx,
                                 self.y + self.action_y + self.vy)
                self.x_root = event.x_root
                self.y_root = event.y_root
        if self.move_notify_time == None or \
           time.time() - self.move_notify_time > 200 * 0.001:
            sstp_event = 'OnEBMouseMove'
            ref = [self.name, int(event.x), int(event.y), self.id]
            header = "NOTIFY SSTP/1.0\r\n" + \
                     "Sender: easyballoon\r\n" + \
                     "Option: nodescript\r\n" + \
                     "Event: %s\r\n" % sstp_event
            for i in range(len(ref)):
                header = header + "Reference%d: %s\r\n" % (i, ref[i])
            header = header + "Charset: Shift_JIS\r\n\r\n"
            self.directsstp_send(header)
            self.move_notify_time = time.time()
        return gtk.TRUE
    def leave_notify(self, window, event):
        self.move_notify_time = None
    def delete(self, window, event):
        return gtk.TRUE
    def destroy(self):
        self.visible = 0
        if self.window:
            self.window.destroy()
            self.window = None
        if self.timeout_id:
            gtk.timeout_remove(self.timeout_id)
            self.timeout_id = None
        gtk.mainquit()
        os._exit(1)

if __name__ == "__main__":
    apply(Balloon, sys.argv[1:])
